"""
supermorecado is an adaptation of mapbox/supermorecado to work with other TMS grid.

"""

__version__ = "0.1.0"

from .burntiles import burnTiles  # noqa
from .densitytiles import densityTiles  # noqa
from .edge_finder import findedges  # noqa
from .uniontiles import unionTiles  # noqa
